/*
 * socket.h
 *
 * Berkley sockets ancestor class.
 *
 * Portable Windows Library
 *
 * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Portable Windows Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
 * All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * $Revision: 29417 $
 * $Author: rjongbloed $
 * $Date: 2013-04-03 11:49:58 +1100 (Wed, 03 Apr 2013) $
 */

///////////////////////////////////////////////////////////////////////////////
// PSocket

  public:
    virtual bool Read(void * ptr, PINDEX len);
    virtual bool Write(const void * buf, PINDEX len);
    ~PSocket();
#if P_HAS_RECVMSG_MSG_ERRQUEUE
  protected:
    virtual PBoolean ConvertOSError(P_INT_PTR libcReturnValue, ErrorGroup group = LastGeneralError);
#endif


//)nd Of File ////////////////////////////////////////////////////////////////
